@@ -30,6 +30,7 @@ from horovod.mxnet.mpi_ops import size, local_size, cross_size, rank, local_rank
 from horovod.mxnet.mpi_ops import mpi_threads_supported, mpi_enabled, mpi_built
 from horovod.mxnet.mpi_ops import gloo_enabled, gloo_built
 from horovod.mxnet.mpi_ops import nccl_built, ddl_built, ccl_built, cuda_built, rocm_built
+from horovod.mxnet.mpi_ops import ProcessSet, global_process_set, add_process_set, remove_process_set
 
 import mxnet as mx
 from collections import OrderedDict, defaultdict
@@ -39,26 +40,29 @@ import warnings
 
 # This is where Horovod's DistributedOptimizer wrapper for MXNet goes
 class DistributedOptimizer(mx.optimizer.Optimizer):
-    def __init__(self, optimizer, compression=Compression.none, gradient_predivide_factor=1.0, num_groups=0):
-        self._compression = compression
+    def __init__(self, optimizer, gradient_predivide_factor=1.0, num_groups=0, process_set=global_process_set):
         if gradient_predivide_factor != 1.0 and rocm_built():
             raise ValueError('gradient_predivide_factor not supported yet with ROCm')
 
         self._optimizer = optimizer
         # Normalizing rescale_grad by Horovod size, which is equivalent to
         # performing average in allreduce, has better performance.
-        self._optimizer.rescale_grad *= (gradient_predivide_factor / size())
+        self._optimizer.rescale_grad *= (gradient_predivide_factor / process_set.size())
         self._gradient_predivide_factor = gradient_predivide_factor
         self._num_groups = num_groups
+        self._process_set = process_set
 
     def __getattr__(self, item):
         return getattr(self._optimizer, item)
 
+    def create_state(self, index, weight):
+        return self._optimizer.create_state(index, weight)
+
     def create_state_multi_precision(self, index, weight):
         return self._optimizer.create_state_multi_precision(index, weight)
 
     def _do_allreduce(self, index, grad):
-        if size() == 1: return
+        if self._process_set.size() == 1: return
 
         if isinstance(index, (tuple, list)):
             if (self._num_groups > 0):
@@ -66,30 +70,28 @@ class DistributedOptimizer(mx.optimizer.Optimizer):
                 index_split = split_list(index, self._num_groups)
 
                 for i, (grads, indices) in enumerate(zip(grad_split, index_split)):
-                    tensors_compressed, ctxs = zip(*[self._compression.compress(grad) for grad in grads])
-                    grouped_allreduce_(tensors=tensors_compressed, average=False,
-                                       name="{}:{}".format(names[0], names[-1]), priority=-i,
-                                       prescale_factor=1.0 / self._gradient_predivide_factor)
-                    grads = [self._compression.decompress(t, ctx) for t, ctx in zip(tensors_compressed, ctxs)]
+                    grouped_allreduce_(tensors=grads, average=False, name="{}:{}".format(indices[0], indices[-1]), priority=-i,
+                                       prescale_factor=1.0 / self._gradient_predivide_factor,
+                                       process_set=self._process_set)
             else:
               for i in range(len(index)):
-                  tensor_compressed, ctx = self._compression.compress(grad[i])
-                  allreduce_(tensor_compressed, average=False,
+                  allreduce_(grad[i], average=False,
                              name=str(index[i]), priority=-i,
-                             prescale_factor=1.0 / self._gradient_predivide_factor)
-                  grad[i] = self._compression.decompress(tensor_compressed, ctx)
+                             prescale_factor=1.0 / self._gradient_predivide_factor,
+                             process_set=self._process_set)
         else:
-            tensor_compressed, ctx = self._compression.compress(grad)
-            allreduce_(tensor_compressed, average=False, name=str(index),
-                       prescale_factor=1.0 / self._gradient_predivide_factor)
-            grad = self._compression.decompress(tensor_compressed, ctx)
+            allreduce_(grad, average=False, name=str(index),
+                       prescale_factor=1.0 / self._gradient_predivide_factor,
+                       process_set=self._process_set)
 
     def update(self, index, weight, grad, state):
-        self._do_allreduce(index, grad)
+        if self._process_set.included():
+            self._do_allreduce(index, grad)
         self._optimizer.update(index, weight, grad, state)
 
     def update_multi_precision(self, index, weight, grad, state):
-        self._do_allreduce(index, grad)
+        if self._process_set.included():
+            self._do_allreduce(index, grad)
         self._optimizer.update_multi_precision(index, weight, grad, state)
 
     def set_learning_rate(self, lr):
@@ -129,8 +131,9 @@ class DistributedTrainer(mx.gluon.Trainer):
     def __init__(self, params, optimizer, optimizer_params=None,
                  compression=Compression.none,
                  gradient_predivide_factor=1.0, prefix=None,
-                 num_groups=0):
+                 num_groups=0, process_set=global_process_set):
         self._compression = compression
+        self._process_set = process_set
 
         if gradient_predivide_factor != 1.0 and rocm_built():
             raise ValueError('gradient_predivide_factor not supported yet with ROCm')
@@ -153,22 +156,28 @@ class DistributedTrainer(mx.gluon.Trainer):
         # _scale is used to check and set rescale_grad for optimizer in Trainer.step()
         # function. Normalizing it by Horovod size, which is equivalent to performing
         # average in allreduce, has better performance. 
-        self._scale *= (gradient_predivide_factor / size())
+        self._scale *= (gradient_predivide_factor / process_set.size())
         self._gradient_predivide_factor = gradient_predivide_factor
         assert prefix is None or isinstance(prefix, str)
         self._prefix = prefix if prefix else ""
         self._num_groups = num_groups
 
     def _allreduce_grads(self):
-        if size() == 1: return
+        if self._process_set.size() == 1: return
+        if not self._process_set.included(): return
 
         if (self._num_groups > 0):
             grads = []
             names = []
+            tensors_compressed = []
+            ctxs = []
 
             for i, param in enumerate(self._params):
                 if param.grad_req != 'null':
-                    grads.append(param.list_grad()[0])
+                    tensor_compressed, ctx = self._compression.compress(param.list_grad()[0])
+                    grads.append(tensor_compressed)
+                    tensors_compressed.append(tensor_compressed)
+                    ctxs.append(ctx)
                     names.append(self._prefix + str(i))
 
             grads_split = split_list(grads, self._num_groups)
@@ -182,10 +191,14 @@ class DistributedTrainer(mx.gluon.Trainer):
 
                 for entries in entries_by_dtype.values():
                     grads, names = zip(*entries)
-                    tensors_compressed, ctxs = zip(*[self._compression.compress(grad) for grad in grads])
-                    grouped_allreduce_(tensors=tensors_compressed, average=False, name="{}:{}".format(names[0], names[-1]), priority=-i,
-                                       prescale_factor=1.0 / self._gradient_predivide_factor)
-                    grads = [self._compression.decompress(t, ctx) for t, ctx in zip(tensors_compressed, ctxs)]
+                    grouped_allreduce_(tensors=grads, average=False, name="{}:{}".format(names[0], names[-1]), priority=-i,
+                                       prescale_factor=1.0 / self._gradient_predivide_factor,
+                                       process_set=self._process_set)
+
+            if self._compression != Compression.none:
+                for i, param in enumerate(self._params):
+                    if param.grad_req != 'null':
+                        param.list_grad()[0][:] = self._compression.decompress(tensors_compressed.pop(0), ctxs.pop(0))
         else:
             # In MXNet 2.0, param.name is no longer unique.
             # Meanwhile, since horovod requires Python 3.6, there is no need to sort
@@ -195,9 +208,11 @@ class DistributedTrainer(mx.gluon.Trainer):
                     tensor_compressed, ctx = self._compression.compress(param.list_grad()[0])
                     allreduce_(tensor_compressed, average=False,
                                name=self._prefix + str(i), priority=-i,
-                               prescale_factor=1.0 / self._gradient_predivide_factor)
-                    param._grad[0][:] = self._compression.decompress(tensor_compressed, ctx)
+                               prescale_factor=1.0 / self._gradient_predivide_factor,
+                               process_set=self._process_set)
 
+                    if self._compression != Compression.none:
+                        param.list_grad()[0][:] = self._compression.decompress(tensor_compressed, ctx)
 
 # Wrapper to inject Horovod broadcast after parameter initialization
 def _append_broadcast_init(param, root_rank, name):
