@@ -20,6 +20,10 @@
 #include "cuda_util.h"
 #include "mpi_ops.h"
 
+#if MXNET_MAJOR >= 2 || MXNET_ASYNC_GPU_ENGINE_SUPPORTED
+#define MXNET_ASYNC_GPU_ENGINE_SUPPORTED 1
+#endif
+
 namespace horovod {
 namespace mxnet {
 
@@ -72,7 +76,66 @@ bool IsTensorOnCPU(NDArray* tensor) {
   return tensor->ctx().dev_mask() == cpu::kDevMask;
 }
 
+#if HAVE_CUDA
+class MXReadyEvent : public common::ReadyEvent {
+public:
+  MXReadyEvent(gpuEvent_t event) : event_(event) {};
+  bool Ready() const override {
+      HVD_GPU_CHECK(gpuEventSynchronize(event_));
+      return true;
+  };
+  gpuEvent_t event() const override {
+    return event_;
+  }
+
+private:
+  gpuEvent_t event_;
+};
+#endif
+
+ReadyEventList FormReadyEventList(NDArray* input, NDArray* output) {
+  ReadyEventList ready_event_list;
+
+#if HAVE_CUDA && MXNET_ASYNC_GPU_ENGINE_SUPPORTED
+  // Get events from input tensor writers
+  {
+    auto& sync_obj = input->var()->sync_object;
+    std::lock_guard<std::mutex> l(sync_obj.mutex);
+    if (!sync_obj.writer_event.empty()) {
+      auto ev = sync_obj.writer_event[0].event.lock();
+      if (ev) {
+        ready_event_list.AddReadyEvent(std::make_shared<MXReadyEvent>(*ev));
+      }
+    }
+  }
+
+  // Get events from output tensor reader and writers
+  {
+    auto& sync_obj = output->var()->sync_object;
+    std::lock_guard<std::mutex> l(sync_obj.mutex);
+    for (auto& cuda_event : sync_obj.reader_events) {
+      auto ev = cuda_event.event.lock();
+      if (ev) {
+        ready_event_list.AddReadyEvent(std::make_shared<MXReadyEvent>(*ev));
+      }
+    }
+    if (!sync_obj.writer_event.empty()) {
+      auto ev = sync_obj.writer_event[0].event.lock();
+      if (ev) {
+        ready_event_list.AddReadyEvent(std::make_shared<MXReadyEvent>(*ev));
+      }
+    }
+  }
+#endif
+  return ready_event_list;
+}
+
+#if MXNET_ASYNC_GPU_ENGINE_SUPPORTED
+void DoHorovodOperation(void*, void* on_start_ptr, void* on_complete_ptr, void* param) {
+  auto on_start = *static_cast<CallbackOnStart*>(on_start_ptr);
+#else
 void DoHorovodOperation(void*, void* on_complete_ptr, void* param) {
+#endif
   ThrowIfError(common::CheckInitialized());
 
   auto on_complete = *static_cast<CallbackOnComplete*>(on_complete_ptr);
@@ -83,20 +146,25 @@ void DoHorovodOperation(void*, void* on_complete_ptr, void* param) {
   auto prescale_factor = ops_param->prescale_factor;
   auto postscale_factor = ops_param->postscale_factor;
   auto num_tensors = ops_param->input_tensors.size();
+  auto process_set_id = ops_param->process_set_id;
 
   std::vector<std::shared_ptr<Tensor>> hvd_tensors;
   std::vector<std::shared_ptr<OpContext>> hvd_contexts;
   std::vector<StatusCallback> callbacks;
-  std::vector<std::shared_ptr<ReadyEvent>> ready_events(num_tensors); // default initialization to nullptr
+  std::vector<ReadyEventList> ready_event_lists;
   hvd_tensors.reserve(num_tensors);
   hvd_contexts.reserve(num_tensors);
+  ready_event_lists.reserve(num_tensors);
   callbacks.reserve(num_tensors);
 
   auto callback_mutex = std::make_shared<std::mutex>();
   for (int i = 0; i < num_tensors; ++i) {
     auto input_tensor = ops_param->input_tensors[i].get();
+    auto output_tensor = ops_param->output_tensors[i].get();
     auto output = ops_param->outputs[i];
 
+    ready_event_lists.emplace_back(FormReadyEventList(input_tensor, output_tensor));
+
     hvd_tensors.emplace_back(std::make_shared<MXTensor>(input_tensor));
     if (TensorUtil::GetDevice(input_tensor) != device) {
       throw std::logic_error("Tensors in list must be on same device.");
@@ -106,8 +174,61 @@ void DoHorovodOperation(void*, void* on_complete_ptr, void* param) {
       ctx->AddOutput(ops_param->received_splits_tensor.get());
     }
     hvd_contexts.push_back(ctx);
-    callbacks.emplace_back([on_complete, ops_param, callback_mutex](const Status& status) {
-      // Must only invoke callback on last tensor to prevent premature deletion of
+    callbacks.emplace_back([on_complete, ops_param, callback_mutex, i](const Status& status) {
+      auto input_tensor = ops_param->input_tensors[i].get();
+      auto output_tensor = ops_param->output_tensors[i].get();
+#if HAVE_CUDA
+      auto hvd_event = status.event;
+      if (hvd_event.event) {
+#if MXNET_ASYNC_GPU_ENGINE_SUPPORTED
+        auto async_engine_enabled = dmlc::GetEnv("MXNET_ASYNC_GPU_ENGINE", false);
+        if (async_engine_enabled) {
+          {
+            auto &sync_obj = input_tensor->var()->sync_object;
+            std::lock_guard<std::mutex> l(sync_obj.mutex);
+            // If some reader event is already recorded on the same stream,
+            // we want to replace ourselves by it
+            int i;
+            for (i = 0; i < sync_obj.reader_events.size(); ++i) {
+              auto stream = sync_obj.reader_events[i].stream;
+              if (stream == hvd_event.stream) {
+                sync_obj.reader_events[i].event = hvd_event.event;
+                sync_obj.reader_events[i].pool_index = hvd_event.event_idx;
+                break;
+              }
+            }
+            if (i == sync_obj.reader_events.size()) {
+              sync_obj.reader_events.push_back({hvd_event.event, hvd_event.stream, hvd_event.event_idx});
+            }
+          }
+
+          {
+            auto &sync_obj = output_tensor->var()->sync_object;
+            std::lock_guard<std::mutex> l(sync_obj.mutex);
+            sync_obj.reader_events.clear();
+            sync_obj.writer_event.clear();
+            sync_obj.writer_event.push_back({hvd_event.event, hvd_event.stream, hvd_event.event_idx});
+          }
+
+          if (ops_param->received_splits_tensor) {
+            {
+              auto &sync_obj = ops_param->received_splits_tensor.get()->var()->sync_object;
+              std::lock_guard<std::mutex> l(sync_obj.mutex);
+              sync_obj.reader_events.clear();
+              sync_obj.writer_event.clear();
+              sync_obj.writer_event.push_back({hvd_event.event, hvd_event.stream, hvd_event.event_idx});
+            }
+          }
+        } else {
+          HVD_GPU_CHECK(gpuEventSynchronize(*(hvd_event.event)));
+        }
+#else
+        HVD_GPU_CHECK(gpuEventSynchronize(*(hvd_event.event)));
+#endif
+      }
+#endif
+
+      // Must only invoke MXNet callback on last tensor to prevent premature deletion of
       // shared ops_param structure. Guard logic is here instead of within DeleteMpiOpsParam
       // function as on_complete can only be invoked once due to MXNet internally
       // pairing up the engine op completion callback with DeleteMpiOpsParam.
@@ -130,13 +251,15 @@ void DoHorovodOperation(void*, void* on_complete_ptr, void* param) {
       }
 
       enqueue_result = EnqueueTensorAllreduces(
-          hvd_contexts, hvd_tensors, hvd_outputs, ready_events, ops_param->op_names, device,
-          callbacks, (average) ? ReduceOp::AVERAGE : ReduceOp::SUM, prescale_factor, postscale_factor);
+          hvd_contexts, hvd_tensors, hvd_outputs, ready_event_lists,
+          ops_param->op_names, device, callbacks,
+          (average) ? ReduceOp::AVERAGE : ReduceOp::SUM, prescale_factor,
+          postscale_factor, process_set_id);
       break;
     case OperationType::ALLGATHER:
       enqueue_result = EnqueueTensorAllgather(
-          hvd_contexts[0], hvd_tensors[0], ready_events[0], ops_param->op_names[0], device,
-          callbacks[0]);
+          hvd_contexts[0], hvd_tensors[0], ready_event_lists[0],
+          ops_param->op_names[0], device, callbacks[0], process_set_id);
       break;
     case OperationType::BROADCAST:
       if (horovod_rank() != ops_param->root_rank) {
@@ -147,15 +270,18 @@ void DoHorovodOperation(void*, void* on_complete_ptr, void* param) {
 
       enqueue_result = EnqueueTensorBroadcast(
           hvd_contexts[0], hvd_tensors[0], hvd_outputs[0], ops_param->root_rank,
-          ready_events[0], ops_param->op_names[0], device,
-          callbacks[0]);
+          ready_event_lists[0], ops_param->op_names[0], device, callbacks[0],
+          process_set_id);
       break;
     case OperationType::ALLTOALL:
     {
+#if MXNET_ASYNC_GPU_ENGINE_SUPPORTED
+      on_start(); // Need to call on_start to sync on possible D2H copy of splits tensor.
+#endif
       auto hvd_splits = std::make_shared<MXTensor>(ops_param->splits_tensor.get());
       enqueue_result = EnqueueTensorAlltoall(
-          hvd_contexts[0], hvd_tensors[0], hvd_splits, ready_events[0], ops_param->op_names[0],
-          device, callbacks[0]);
+          hvd_contexts[0], hvd_tensors[0], hvd_splits, ready_event_lists[0],
+          ops_param->op_names[0], device, callbacks[0], process_set_id);
       break;
     }
     default:
@@ -167,7 +293,9 @@ void DoHorovodOperation(void*, void* on_complete_ptr, void* param) {
 
 inline void PushHorovodOperation(OperationType op_type, NDArray* const * inputs,
                                  NDArray* const * outputs, const char* name,
-                                 int priority, int num_tensors, int root_rank = -1,
+                                 int priority, int num_tensors,
+                                 int process_set_id = 0,
+                                 int root_rank = -1,
                                  bool average = true,
                                  NDArray* splits = nullptr,
                                  NDArray* output_received_splits = nullptr,
@@ -240,9 +368,11 @@ inline void PushHorovodOperation(OperationType op_type, NDArray* const * inputs,
     received_splits_tensor = std::make_shared<NDArray>(*output_received_splits);
   }
 
-  auto ops_param = CreateMpiOpsParam(std::move(input_copies), std::move(output_copies),
-    std::move(outputs_vec), cpu_input_tensors, cpu_output_tensors, op_type, std::move(op_names), root_rank,
-    average, splits_tensor, received_splits_tensor, prescale_factor, postscale_factor);
+  auto ops_param = CreateMpiOpsParam(
+      std::move(input_copies), std::move(output_copies), std::move(outputs_vec),
+      cpu_input_tensors, cpu_output_tensors, op_type, std::move(op_names),
+      root_rank, average, splits_tensor, received_splits_tensor,
+      prescale_factor, postscale_factor, process_set_id);
 
   // Not in-place
   if (!inplace) {
@@ -266,7 +396,12 @@ inline void PushHorovodOperation(OperationType op_type, NDArray* const * inputs,
   }
 }
 #if HAVE_CUDA
+#if MXNET_ASYNC_GPU_ENGINE_SUPPORTED
+void DoHorovodOperationCudaOnCPU(void*, void* on_start_ptr, void* on_complete_ptr, void* param) {
+  auto on_start = *static_cast<CallbackOnComplete*>(on_start_ptr);
+#else
 void DoHorovodOperationCudaOnCPU(void*, void* on_complete_ptr, void* param) {
+#endif
   ThrowIfError(common::CheckInitialized());
 
   auto on_complete = *static_cast<CallbackOnComplete*>(on_complete_ptr);
@@ -276,13 +411,15 @@ void DoHorovodOperationCudaOnCPU(void*, void* on_complete_ptr, void* param) {
   auto prescale_factor = ops_param->prescale_factor;
   auto postscale_factor = ops_param->postscale_factor;
   auto num_tensors = ops_param->cpu_input_tensors.size();
+  auto process_set_id = ops_param->process_set_id;
 
   std::vector<std::shared_ptr<Tensor>> hvd_cpu_buffers;
   std::vector<std::shared_ptr<OpContext>> hvd_contexts;
   std::vector<StatusCallback> callbacks;
-  std::vector<std::shared_ptr<ReadyEvent>> ready_events(num_tensors); // default initialization to nullptr
+  std::vector<ReadyEventList> ready_event_lists;
   hvd_cpu_buffers.reserve(num_tensors);
   hvd_contexts.reserve(num_tensors);
+  ready_event_lists.reserve(num_tensors);
   callbacks.reserve(num_tensors);
 
   auto callback_mutex = std::make_shared<std::mutex>();
@@ -290,6 +427,8 @@ void DoHorovodOperationCudaOnCPU(void*, void* on_complete_ptr, void* param) {
     auto input = ops_param->cpu_input_tensors[i].get();
     auto output = ops_param->cpu_output_tensors[i].get();
 
+    ready_event_lists.emplace_back(FormReadyEventList(input, output));
+
     hvd_cpu_buffers.emplace_back(std::make_shared<MXTensor>(input));
     if (TensorUtil::GetDevice(input) != device) {
       throw std::logic_error("Tensors in list must be on same device.");
@@ -315,32 +454,34 @@ void DoHorovodOperationCudaOnCPU(void*, void* on_complete_ptr, void* param) {
 
   Status enqueue_result;
   switch (ops_param->op_type) {
-    case OperationType::ALLREDUCE:
-      enqueue_result = EnqueueTensorAllreduces(
-          hvd_contexts, hvd_cpu_buffers, hvd_cpu_buffers, ready_events, ops_param->op_names, device,
-          callbacks, (average) ? ReduceOp::AVERAGE : ReduceOp::SUM, prescale_factor, postscale_factor);
-      break;
-    case OperationType::ALLGATHER:
-      enqueue_result = EnqueueTensorAllgather(
-          hvd_contexts[0], hvd_cpu_buffers[0], ready_events[0], ops_param->op_names[0], device,
-          callbacks[0]);
-      break;
-    case OperationType::BROADCAST:
-      enqueue_result = EnqueueTensorBroadcast(
-          hvd_contexts[0], hvd_cpu_buffers[0], hvd_cpu_buffers[0], ops_param->root_rank,
-          ready_events[0], ops_param->op_names[0], device,
-          callbacks[0]);
-      break;
-    case OperationType::ALLTOALL:
-    {
-      auto hvd_splits = std::make_shared<MXTensor>(ops_param->splits_tensor.get());
-      enqueue_result = EnqueueTensorAlltoall(
-          hvd_contexts[0], hvd_cpu_buffers[0], hvd_splits, ready_events[0], ops_param->op_names[0],
-          device, callbacks[0]);
-      break;
-    }
-    default:
-      throw std::logic_error("Unsupported Horovod operation type.");
+  case OperationType::ALLREDUCE:
+    enqueue_result = EnqueueTensorAllreduces(
+        hvd_contexts, hvd_cpu_buffers, hvd_cpu_buffers, ready_event_lists,
+        ops_param->op_names, device, callbacks,
+        (average) ? ReduceOp::AVERAGE : ReduceOp::SUM, prescale_factor,
+        postscale_factor, process_set_id);
+    break;
+  case OperationType::ALLGATHER:
+    enqueue_result = EnqueueTensorAllgather(
+        hvd_contexts[0], hvd_cpu_buffers[0], ready_event_lists[0],
+        ops_param->op_names[0], device, callbacks[0], process_set_id);
+    break;
+  case OperationType::BROADCAST:
+    enqueue_result = EnqueueTensorBroadcast(
+        hvd_contexts[0], hvd_cpu_buffers[0], hvd_cpu_buffers[0],
+        ops_param->root_rank, ready_event_lists[0], ops_param->op_names[0],
+        device, callbacks[0], process_set_id);
+    break;
+  case OperationType::ALLTOALL: {
+    auto hvd_splits =
+        std::make_shared<MXTensor>(ops_param->splits_tensor.get());
+    enqueue_result = EnqueueTensorAlltoall(
+        hvd_contexts[0], hvd_cpu_buffers[0], hvd_splits, ready_event_lists[0],
+        ops_param->op_names[0], device, callbacks[0], process_set_id);
+    break;
+  }
+  default:
+    throw std::logic_error("Unsupported Horovod operation type.");
   }
 
   ThrowIfError(enqueue_result);
@@ -348,13 +489,14 @@ void DoHorovodOperationCudaOnCPU(void*, void* on_complete_ptr, void* param) {
 
 inline void PushHorovodOperationCudaOnCPU(OperationType op_type, NDArray* const * inputs,
                                           NDArray* const * outputs, const char* name,
-                                          int priority, int num_tensors, int root_rank = -1,
+                                          int priority, int num_tensors,
+                                          int process_set_id = 0,
+                                          int root_rank = -1,
                                           bool average = true,
                                           NDArray* splits = nullptr,
                                           NDArray* output_received_splits = nullptr,
                                           double prescale_factor = 1.0,
                                           double postscale_factor = 1.0) {
-
   auto op_type_name = GetOpTypeName(op_type);
 
   std::vector<std::shared_ptr<NDArray>> input_copies; //empty
@@ -404,9 +546,11 @@ inline void PushHorovodOperationCudaOnCPU(OperationType op_type, NDArray* const
     received_splits_tensor = std::make_shared<NDArray>(*output_received_splits);
   }
 
-  auto ops_param = CreateMpiOpsParam(std::move(input_copies), std::move(output_copies),
-    std::move(outputs_vec), cpu_input_tensors, cpu_output_tensors, op_type, std::move(op_names), root_rank,
-    average, splits_tensor, received_splits_tensor, prescale_factor, postscale_factor);
+  auto ops_param = CreateMpiOpsParam(
+      std::move(input_copies), std::move(output_copies), std::move(outputs_vec),
+      cpu_input_tensors, cpu_output_tensors, op_type, std::move(op_names),
+      root_rank, average, splits_tensor, received_splits_tensor,
+      prescale_factor, postscale_factor, process_set_id);
 
   std::vector<void*> cpu_input_vars;
   std::vector<void*> cpu_output_vars;
@@ -458,7 +602,8 @@ extern "C" int horovod_mxnet_allreduce_async(NDArray* const * inputs,
                                              int priority,
                                              double prescale_factor,
                                              double postscale_factor,
-                                             int num_tensors) {
+                                             int num_tensors,
+                                             int process_set_id) {
   MX_API_BEGIN();
 
 #if HAVE_ROCM
@@ -470,16 +615,19 @@ extern "C" int horovod_mxnet_allreduce_async(NDArray* const * inputs,
 
 #if HAVE_CUDA && !HOROVOD_GPU_ALLREDUCE
   if (IsTensorOnCPU(inputs[0]) && IsTensorOnCPU(outputs[0])) {
-    PushHorovodOperation(OperationType::ALLREDUCE, inputs, outputs,
-                         name, priority, num_tensors, -1, average, nullptr, nullptr, prescale_factor, postscale_factor);
+    PushHorovodOperation(OperationType::ALLREDUCE, inputs, outputs, name,
+                         priority, num_tensors, process_set_id, -1, average,
+                         nullptr, nullptr, prescale_factor, postscale_factor);
   } else {
     PushHorovodOperationCudaOnCPU(OperationType::ALLREDUCE, inputs, outputs,
-                                  name, priority, num_tensors, -1, average, nullptr, nullptr, prescale_factor, postscale_factor);
+                                  name, priority, num_tensors, process_set_id,
+                                  -1, average, nullptr, nullptr,
+                                  prescale_factor, postscale_factor);
   }
 #else
-  PushHorovodOperation(OperationType::ALLREDUCE, inputs, outputs,
-                       name, priority, num_tensors, -1, average, nullptr,
-                       nullptr, prescale_factor, postscale_factor);
+  PushHorovodOperation(OperationType::ALLREDUCE, inputs, outputs, name,
+                       priority, num_tensors, process_set_id, -1, average,
+                       nullptr, nullptr, prescale_factor, postscale_factor);
 #endif
 
 #if HAVE_ROCM
@@ -495,20 +643,21 @@ extern "C" int horovod_mxnet_allreduce_async(NDArray* const * inputs,
 
 extern "C" int horovod_mxnet_allgather_async(NDArray* input,
                                              NDArray* output,
-                                             const char* name, int priority) {
+                                             const char* name, int priority,
+                                             int process_set_id) {
   MX_API_BEGIN();
 
 #if HAVE_CUDA && !HOROVOD_GPU_ALLGATHER
   if (IsTensorOnCPU(input) && IsTensorOnCPU(output)) {
     PushHorovodOperation(OperationType::ALLGATHER, &input, &output,
-                         name, priority, 1);
+                         name, priority, 1, process_set_id);
   } else {
     PushHorovodOperationCudaOnCPU(OperationType::ALLGATHER, &input, &output,
-                                  name, priority, 1);
+                                  name, priority, 1, process_set_id);
   }
 #else
   PushHorovodOperation(OperationType::ALLGATHER, &input, &output,
-                       name, priority, 1);
+                       name, priority, 1, process_set_id);
 #endif
 
   MX_API_END();
@@ -517,21 +666,22 @@ extern "C" int horovod_mxnet_allgather_async(NDArray* input,
 extern "C" int horovod_mxnet_broadcast_async(NDArray* input,
                                              NDArray* output,
                                              const char* name, int root_rank,
-                                             int priority) {
+                                             int priority,
+                                             int process_set_id) {
   MX_API_BEGIN();
 
 #if HAVE_CUDA && !HOROVOD_GPU_BROADCAST
   if (IsTensorOnCPU(input) && IsTensorOnCPU(output)) {
     PushHorovodOperation(OperationType::BROADCAST, &input, &output,
-                         name, priority, 1, root_rank);
+                         name, priority, 1, process_set_id, root_rank);
 
   } else {
     PushHorovodOperationCudaOnCPU(OperationType::BROADCAST, &input, &output,
-                                  name, priority, 1, root_rank);
+                                  name, priority, 1, process_set_id, root_rank);
   }
 #else
-  PushHorovodOperation(OperationType::BROADCAST, &input, &output,
-                       name, priority, 1, root_rank);
+  PushHorovodOperation(OperationType::BROADCAST, &input, &output, name,
+                       priority, 1, process_set_id, root_rank);
 #endif
 
   MX_API_END();
@@ -542,23 +692,25 @@ extern "C" int horovod_mxnet_alltoall_async(NDArray* input,
                                             const char* name,
                                             NDArray* splits,
                                             NDArray* output_received_splits,
-                                            int priority) {
+                                            int priority,
+                                            int process_set_id) {
   MX_API_BEGIN();
 
 #if HAVE_CUDA && !HOROVOD_GPU_ALLTOALL
   if (IsTensorOnCPU(input) && IsTensorOnCPU(output)) {
     PushHorovodOperation(OperationType::ALLTOALL, &input, &output, name,
-                         priority, 1, -1, false, splits,
+                         priority, 1, process_set_id, -1, false, splits,
                          output_received_splits);
 
   } else {
     PushHorovodOperationCudaOnCPU(OperationType::ALLTOALL, &input, &output,
-                                  name, priority, 1, -1, false, splits,
-                                  output_received_splits);
+                                  name, priority, 1, process_set_id, -1, false,
+                                  splits, output_received_splits);
   }
 #else
   PushHorovodOperation(OperationType::ALLTOALL, &input, &output, name, priority,
-                       1, -1, false, splits, output_received_splits);
+                       1, process_set_id, -1, false, splits,
+                       output_received_splits);
 #endif
 
   MX_API_END();
