@@ -14,6 +14,7 @@
 # ==============================================================================
 
 import base64
+import pickle
 import cloudpickle
 
 
@@ -23,6 +24,6 @@ def loads_base64(encoded):
 
 
 def dumps_base64(obj, to_ascii=True):
-    serialized = cloudpickle.dumps(obj)
+    serialized = cloudpickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL)
     encoded = base64.b64encode(serialized)
     return encoded.decode('ascii') if to_ascii else encoded
