@@ -15,20 +15,48 @@
 
 import io
 import logging
+import random
 import threading
-
+import time
 from collections import defaultdict
 
 from horovod.runner.common.util import safe_shell_exec
 from horovod.runner.elastic.worker import HostUpdateResult
 
+# The default lower bound for cooldown period. If a range is provided,
+# the provided lower limit must be at or above this lower bound
+DEFAULT_COOLDOWN_LOWER_LIMIT_SECONDS = 1
+# The default upper bound for cooldown period. If a range is provided,
+# the provided upper limit must be at or below this upper bound
+DEFAULT_COOLDOWN_UPPER_LIMIT_SECONDS = 1 * 60 * 60
 
 class HostState(object):
-    def __init__(self):
+
+    def __init__(self, cooldown_range=None):
         self._event = threading.Event()
 
-        # TODO(travis): blacklisted hosts should have a timeout period that increases with each failure
         self._blacklisted = False
+        self._blacklist_count = 0
+        if cooldown_range:
+            HostState._validate_cooldown_range(cooldown_range)
+            self._cooldown_lower_limit, self._cooldown_upper_limit = cooldown_range
+        else:
+            self._cooldown_lower_limit = -1
+            self._cooldown_upper_limit = -1
+        self._cooldown_period_end_ts = 0
+
+    @staticmethod
+    def _validate_cooldown_range(cooldown_range):
+        cooldown_lower_limit, cooldown_upper_limit = cooldown_range
+
+        if (cooldown_lower_limit < DEFAULT_COOLDOWN_LOWER_LIMIT_SECONDS):
+            raise ValueError(f"Provided cooldown lower limit: {cooldown_lower_limit} \
+                             cannot be lower than default cooldown lower limit: {DEFAULT_COOLDOWN_LOWER_LIMIT_SECONDS}")
+
+
+        if (cooldown_upper_limit > DEFAULT_COOLDOWN_UPPER_LIMIT_SECONDS):
+            raise ValueError(f"Provided cooldown upper limit: {cooldown_upper_limit} \
+                             cannot be higher than default cooldown upper limit: {DEFAULT_COOLDOWN_UPPER_LIMIT_SECONDS}")
 
     def get_event(self):
         if self._event.is_set():
@@ -39,13 +67,48 @@ class HostState(object):
     def set_event(self):
         self._event.set()
 
+    def _in_cooldown_period(self, current_time):
+        return self._cooldown_period_end_ts > current_time
+
+
+    def _set_cooldown_period(self, current_time):
+        if self._cooldown_lower_limit == -1 or self._cooldown_upper_limit == -1:
+            return
+        self._blacklist_count += 1
+
+        cooldown_delay = self._cooldown_lower_limit * (1 << self._blacklist_count) + (random.uniform(0,1) * self._cooldown_lower_limit)
+        logging.debug(f"{self._blacklist_count}:{self._cooldown_period_end_ts} cooldown_delay: {cooldown_delay}")
+        # We need to ensure that the cooldown upper limit is the upper bound of the delay
+        cooldown_delta_seconds = max(self._cooldown_lower_limit, min(self._cooldown_upper_limit, cooldown_delay))
+
+        self._cooldown_period_end_ts = current_time + cooldown_delta_seconds
+        logging.debug(f"cooldown delta seconds: {cooldown_delta_seconds}")
+
     def blacklist(self):
+        """Moves this host to a blacklist, and starts the cooldown period."""
         self._blacklisted = True
+        now = time.time()
+        if self._in_cooldown_period(now):
+            return
+        self._set_cooldown_period(now)
         self.set_event()
 
+    def whitelist(self):
+        """Ends the cooldown period and moves this host out of blacklist."""
+        self._cooldown_period_end_ts = 0
+        self._blacklisted = False
+
     def is_blacklisted(self):
+        """Checks if the host is in the blacklist."""
         return self._blacklisted
 
+    def is_resurrected(self):
+        """Checks if host is in an expired cooldown period."""
+        if self._cooldown_period_end_ts > 0:
+            return not self._in_cooldown_period(time.time())
+        return False
+
+
 
 class DiscoveredHosts(object):
     def __init__(self, host_slots, host_assignment_order):
@@ -76,15 +139,17 @@ class DiscoveredHosts(object):
                                        if not hosts_state[host].is_blacklisted()]
         return self
 
+    def __str__(self):
+        return f"slots: {self._host_slots} order: {self._host_assignment_order}"
+
 
 class HostManager(object):
-    def __init__(self, discovery):
+    def __init__(self, discovery, cooldown_range=None):
         self._current_hosts = DiscoveredHosts(host_slots={}, host_assignment_order=[])
-        self._hosts_state = defaultdict(HostState)
+        self._hosts_state = defaultdict(lambda: HostState(cooldown_range))
         self._discovery = discovery
 
     def update_available_hosts(self):
-        # TODO(travis): also check for hosts removed from the blacklist in the future
         def check_update(cur_host_slots, prev_host_slots):
             res = HostUpdateResult.no_update
 
@@ -103,17 +168,32 @@ class HostManager(object):
                 elif cur_host_slots[h] < prev_host_slots[h]:
                     # h has removed some slots
                     res |=  HostUpdateResult.removed
+                elif self._hosts_state[h].is_resurrected():
+                    res |= HostUpdateResult.added
             return res
 
         prev_host_slots = self._current_hosts.host_slots
         prev_host_assignment_order = self._current_hosts.host_assignment_order
         host_slots = self._discovery.find_available_hosts_and_slots()
-        if prev_host_slots != host_slots:
-            available_hosts = set([host for host in host_slots.keys() if not self._hosts_state[host].is_blacklisted()])
+
+        def whitelist_all_hosts():
+            for host in host_slots.keys():
+                if self._hosts_state[host].is_resurrected():
+                    self._hosts_state[host].whitelist()
+
+        def has_resurrected_hosts():
+            resurrected_hosts = [host for host in host_slots.keys() if self._hosts_state[host].is_resurrected()]
+            return len(resurrected_hosts) > 0
+
+        if prev_host_slots != host_slots or has_resurrected_hosts():
+            available_hosts = set([host for host in host_slots.keys() \
+                if not (self._hosts_state[host].is_blacklisted() and not self._hosts_state[host].is_resurrected())])
             host_assignment_order = HostManager.order_available_hosts(available_hosts, prev_host_assignment_order)
             self._current_hosts = DiscoveredHosts(host_slots=host_slots,
                                                   host_assignment_order=host_assignment_order)
-            return check_update(self._current_hosts.host_slots, prev_host_slots)
+            host_update_state = check_update(self._current_hosts.host_slots, prev_host_slots)
+            whitelist_all_hosts()
+            return host_update_state
         else:
             return HostUpdateResult.no_update
 
@@ -123,7 +203,7 @@ class HostManager(object):
 
     def blacklist(self, host):
         if not self._hosts_state[host].is_blacklisted():
-            logging.warning('blacklist failing host: {}'.format(host))
+            logging.info('blacklist failing host: {}'.format(host))
         self._hosts_state[host].blacklist()
 
     def is_blacklisted(self, host):
