@@ -12,14 +12,9 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # ==============================================================================
-
-import horovod.spark.common._namedtuple_fix
-
 import numbers
 import time
-
-from distutils.version import LooseVersion
-
+import os
 import numpy as np
 import tensorflow as tf
 
@@ -35,10 +30,7 @@ from horovod.spark.common.estimator import HorovodEstimator, HorovodModel
 from horovod.spark.common.params import EstimatorParams
 from horovod.spark.common.serialization import HorovodParamsWriter, HorovodParamsReader
 from horovod.spark.keras import remote
-from horovod.spark.keras.util import \
-    BARE_KERAS, TF_KERAS, \
-    BareKerasUtil, TFKerasUtil, \
-    is_instance_of_bare_keras_model, is_instance_of_bare_keras_optimizer
+from horovod.spark.keras.util import TFKerasUtil
 
 
 class KerasEstimatorParamsWriter(HorovodParamsWriter):
@@ -75,16 +67,6 @@ class KerasEstimatorParamsReader(HorovodParamsReader):
 
         # In order to deserialize the model, we need to deserialize the custom_objects param
         # first.
-        keras_utils = None
-        if KerasEstimator._keras_pkg_type.name in dict:
-            keras_pkg_type = _param_deserializer_fn(KerasEstimator._keras_pkg_type.name,
-                                                    dict[KerasEstimator._keras_pkg_type.name],
-                                                    None, None)
-            if keras_pkg_type == BARE_KERAS:
-                keras_utils = BareKerasUtil
-            elif keras_pkg_type == TF_KERAS:
-                keras_utils = TFKerasUtil
-
         custom_objects = {}
         if KerasEstimator.custom_objects.name in dict:
             custom_objects = _param_deserializer_fn(KerasEstimator.custom_objects.name,
@@ -92,7 +74,7 @@ class KerasEstimatorParamsReader(HorovodParamsReader):
                                                     None, None)
 
         for key, val in dict.items():
-            dict[key] = _param_deserializer_fn(key, val, keras_utils, custom_objects)
+            dict[key] = _param_deserializer_fn(key, val, TFKerasUtil, custom_objects)
         return dict
 
 
@@ -133,9 +115,10 @@ class KerasEstimator(HorovodEstimator, KerasEstimatorParamsReadable,
         val_batch_size: Number of rows from the DataFrame per batch for validation, if not set, will use batch_size.
         epochs: Number of epochs to train.
         verbose: Verbosity level [0, 2] (default: 1).
-        shuffle_buffer_size: Optional size of in-memory shuffle buffer in rows. Allocating a larger buffer size
-                             increases randomness of shuffling at the cost of more host memory. Defaults to estimating
-                             with an assumption of 4GB of memory per host.
+        shuffle_buffer_size: Optional size of in-memory shuffle buffer in rows (on training data).
+                             Allocating a larger buffer size increases randomness of shuffling at
+                             the cost of more host memory. Defaults to estimating with an assumption
+                             of 4GB of memory per host. Set shuffle_buffer_size=0 would turn off shuffle.
         partitions_per_process: Number of Parquet partitions to assign per worker process from `num_proc` (default: 10).
         run_id: Optional unique ID for this run for organization in the Store. Will be automatically assigned if not
                 provided.
@@ -161,15 +144,18 @@ class KerasEstimator(HorovodEstimator, KerasEstimatorParamsReadable,
         reader_pool_type: Type of worker pool used to parallelize reading data from the dataset.
                           Should be one of ['thread', 'process']. Defaults to 'process'.
         inmemory_cache_all: boolean value. Cache the data in memory for training and validation. Default: False.
+        backend_env: dict to add to the environment of the backend.  Defaults to setting the java heap size to
+                     2G min and max for libhdfs through petastorm
     """
 
     custom_objects = Param(Params._dummy(), 'custom_objects', 'custom objects')
-    _keras_pkg_type = Param(Params._dummy(), '_keras_pkg_type', 'keras package type')
     checkpoint_callback = Param(Params._dummy(), 'checkpoint_callback',
                                 'model checkpointing callback')
     inmemory_cache_all = Param(Params._dummy(), 'inmemory_cache_all',
                                'Cache the data in memory for training and validation.',
                                typeConverter=TypeConverters.toBoolean)
+    backend_env = Param(Params._dummy(), "backend_env",
+                        "dict to add to the environment of the command run on the environment")
 
     @keyword_only
     def __init__(self,
@@ -203,61 +189,37 @@ class KerasEstimator(HorovodEstimator, KerasEstimatorParamsReadable,
                  reader_pool_type=None,
                  label_shapes=None,
                  checkpoint_callback=None,
-                 inmemory_cache_all=False):
+                 inmemory_cache_all=False,
+                 backend_env=None):
 
         super(KerasEstimator, self).__init__()
 
         self._setDefault(optimizer=None,
                          custom_objects={},
-                         _keras_pkg_type=None,
                          checkpoint_callback=None,
-                         inmemory_cache_all=False)
+                         inmemory_cache_all=False,
+                         backend_env={'LIBHDFS_OPTS': '-Xms2048m -Xmx2048m'})
 
         kwargs = self._input_kwargs
         self.setParams(**kwargs)
 
     def _get_keras_utils(self):
-        # This function determines the keras package type of the Estimator based on the passed
-        # optimizer and model and updates _keras_pkg_type parameter.
+        # This function checks the keras package type is tensorflow.keras
 
-        model_type = None
         model = self.getModel()
         if model:
-            if isinstance(model, tf.keras.Model):
-                model_type = TF_KERAS
-            elif is_instance_of_bare_keras_model(model):
-                model_type = BARE_KERAS
-            else:
+            if not isinstance(model, tf.keras.Model):
                 raise ValueError(
-                    "model has to be an instance of tensorflow.keras.Model or keras.Model")
+                    "model has to be an instance of tensorflow.keras.Model")
 
-        optimizer_type = None
         optimizer = self.getOptimizer()
         if optimizer:
             if isinstance(optimizer, str):
-                optimizer_type = None
-            elif isinstance(optimizer, tf.keras.optimizers.Optimizer):
-                optimizer_type = TF_KERAS
-            elif is_instance_of_bare_keras_optimizer(optimizer):
-                optimizer_type = BARE_KERAS
-            else:
-                raise ValueError("invalid optimizer type")
+                pass
+            elif not isinstance(optimizer, tf.keras.optimizers.Optimizer):
+                raise ValueError("optimizer has to be an instance of tensorflow.keras.optimizers.Optimizer")
 
-        types = set([model_type, optimizer_type])
-        types.discard(None)
-
-        if len(types) > 1:
-            raise ValueError('mixed keras and tf.keras values for optimizers and model')
-        elif len(types) == 1:
-            pkg_type = types.pop()
-            super(KerasEstimator, self)._set(_keras_pkg_type=pkg_type)
-
-            if pkg_type == TF_KERAS:
-                return TFKerasUtil
-            elif pkg_type == BARE_KERAS:
-                return BareKerasUtil
-            else:
-                raise ValueError("invalid keras type")
+        return TFKerasUtil
 
     def setCustomObjects(self, value):
         return self._set(custom_objects=value)
@@ -277,6 +239,12 @@ class KerasEstimator(HorovodEstimator, KerasEstimatorParamsReadable,
     def getInMemoryCacheAll(self):
         return self.getOrDefault(self.inmemory_cache_all)
 
+    def setBackendEnv(self, value):
+        self._set(backend_env=value)
+
+    def getBackendEnv(self):
+        return self.getOrDefault(self.backend_env)
+
     def _check_metadata_compatibility(self, metadata):
         input_shapes, output_shapes = self.get_model_shapes()
         util.check_shape_compatibility(metadata,
@@ -307,19 +275,18 @@ class KerasEstimator(HorovodEstimator, KerasEstimatorParamsReadable,
         else:
             serialized_model = self._compile_model(keras_utils)
 
-        # Workaround:
-        # https://stackoverflow.com/questions/50583056/is-there-any-way-to-set-java-opts-for-tensorflow-process/50615570
-        env = {'LIBHDFS_OPTS': '-Xms2048m -Xmx2048m'}
-
         trainer = remote.RemoteTrainer(self, metadata, keras_utils, run_id, dataset_idx)
         handle = backend.run(trainer,
                              args=(serialized_model, train_rows, val_rows, avg_row_size),
-                             env=env)
+                             env=self.getBackendEnv())
         return self._create_model(handle, run_id, metadata)
 
     def _load_model_from_checkpoint(self, run_id):
         store = self.getStore()
-        last_ckpt_path = store.get_checkpoint_path(run_id)
+        last_ckpt_path = os.path.join(store.get_checkpoint_path(run_id), store.get_checkpoint_filename())
+
+        if not store.fs.exists(last_ckpt_path):
+            return None
 
         if self.getVerbose():
             print('Resuming training from last checkpoint: {}'.format(last_ckpt_path))
@@ -420,11 +387,6 @@ class KerasModel(HorovodModel, KerasEstimatorParamsReadable,
     """
 
     custom_objects = Param(Params._dummy(), 'custom_objects', 'custom objects')
-
-    # Setting _keras_pkg_type parameter helps us determine the type of keras package during
-    # deserializing the transformer
-    _keras_pkg_type = Param(Params._dummy(), '_keras_pkg_type', 'keras package type')
-
     _floatx = Param(Params._dummy(), '_floatx', 'keras default float type')
 
     @keyword_only
@@ -458,22 +420,11 @@ class KerasModel(HorovodModel, KerasEstimatorParamsReadable,
         # infer keras package from model
         model = self.getModel()
         if model:
-            if isinstance(model, tf.keras.Model):
-                pkg_type = TF_KERAS
-            elif is_instance_of_bare_keras_model(model):
-                pkg_type = BARE_KERAS
-            else:
+            if not isinstance(model, tf.keras.Model):
                 raise ValueError(
-                    "model has to be an instance of tensorflow.keras.Model or keras.Model")
+                    "model has to be an instance of tensorflow.keras.Model")
 
-            super(KerasModel, self)._set(_keras_pkg_type=pkg_type)
-
-            if pkg_type == TF_KERAS:
-                return TFKerasUtil
-            elif pkg_type == BARE_KERAS:
-                return BareKerasUtil
-            else:
-                raise ValueError("invalid keras type")
+            return TFKerasUtil
 
         raise ValueError("model is not set")
 
@@ -570,4 +521,3 @@ class KerasModel(HorovodModel, KerasEstimatorParamsReadable,
 
         # Use the schema from previous section to construct the final DF with prediction
         return spark0.createDataFrame(pred_rdd, schema=final_output_schema)
-
